#$Id$
#Package Maintainer : support@threatstop.com
#########################################################################
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# United States Patent No. 62539504 (Pending),
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
########################################################################

=head1 SYNOPSIS

 This is a Utility library meant for use with on device installations to support auto configuration
 with minimal lib dependacies (perl only + dig). Currently being used with Vyatta among others.
=cut

package ThreatStop::AutoConfig;
use strict;
use warnings;
use 5.010;    #enable perl 5.10 features.

use Exporter;
our $VERSION = 1.00;
our @ISA     = qw(Exporter);
our @EXPORT  = qw(retrieveDeviceConfig);

# use Switch;
use ThreatStop;
use ThreatStop::STunnel;

# dev
use Data::Dumper;

=head3 retrieveDeviceConfig()



 query for TXT "configuration" records for the given device ID.

 Args :
    hash - install OPTS with all app locations like stunnel, dig, nsupdate
    dnslist - configured name servers to check
    dnsport - port on which to query configured name servers
    device ID - the unique device ID that will be queried
    auto configuration version - gets embedded into the zone name
    TSIG key - used for access to AXFR the zone
    results - hash reference in which to report findings
          ->{status} - set error messages, when return value is 0
          ->{config} - a hash reference to all config key/value pairs

 Returns :
   1 - failure
   0 - success
=cut

sub retrieveDeviceConfig {

    my $OPTS = shift;

    # my $dig     = shift;
    my $dnslist = shift;    # list of dns servers separated by spaces
    my $dnsport = shift;    # DNS port to use.
    my $did     = shift;    # device ID
    my $pcfgver = shift;    # auto config version (indicates the format we'll ask for)
    my $pkey    = shift;    # auto TSIG key, which must come prepended with the key name
    my $results = shift;    # hashref to store error or other information gathered in.
    $results = {} if ( !defined($results) );

    my @dnslist = split( / /, $dnslist );
    my $cmd;

    my $cfgZone = qq($did.v$pcfgver.config.threatstop.local);

    #If no dns list was received
    if ( !@dnslist ) {
        $results->{status} = "DNS servers list is not defined";
        return 1;
    }

    unless ( exists $OPTS->{dig} ) {
        &printMsg('FAIL', "Missing dig application location");
        return 1;
    }

    # update_dig_location($dig);
    return 1 if (stunnel_required($OPTS) == 1);

    #Pass over dns servers list and query zone transfer of the configuration zone
    #
    #!!! Returning in this loop requires calling stunnel_completed($OPTS) to relinquish the use of stunnel
    my $dig_completed = 0;    # Find out if dig fails to connect or to retrieve the CTP zone
    foreach my $dns (@dnslist) {

        $cmd = qq|$OPTS->{dig} +tcp -p $dnsport -t AXFR -y '$did:$pkey' \@$dns $cfgZone 2>&1|;
        printMsg( 'DEBUG', "Using DNS server '$dns'" );
        printMsg( 'DEBUG', "$cmd" );
        my $cmdout = qx($cmd);
        my $rv     = $?;
        printMsg( 'DEBUG', "dig command exited with [$rv]" );

        # Dig returns 0 on connection reset, which can happen if starting stunnel failed
        if ( $rv != 0 || $cmdout =~ /communications error/ ) {

            printMsg( 'DEBUG', "non-zero exit status from dig; moving on to next DNS server" );
            next;
        }

        $dig_completed = 1;

        # note; for many failure types, "dig" will exit with status 0
        # for AXFR, we're only going to consider the transfer a success
        # if the results contain the line :
        # XFR size: 43 records (messages 3, bytes 2100)
        # and the number of records is of a reasonable minimum

        my $minTXTs = 5;    # we don't want to assume exactly how many config
                            # records, but if we don't get at least this many
                            # total records, we won't consider this a good pull
                            # (the core has at least this many)
        if ( $cmdout =~ /XFR size: (\d+) records/s && $1 > $minTXTs ) {

            printMsg( 'DEBUG', "got good zone transfer with $1 TXT records from '$dns'" );

            my @lines = split /\n/, $cmdout;
            my @txtLines   = grep /\s+IN\s+TXT\s+/, @lines;
            my $sumTXT     = '';                              # build up for checksum
            my @txtRecords = ();
            foreach my $line (@txtLines) {

                # only need payload part of the TXT record
                my @lineParts = split( /\s+/, $line, 5 );
                push @txtRecords, $lineParts[4];
            }

            # sort by TXT records by the key part
            foreach my $txtRecord (
                sort {
                    my $akey = ( split( /=/, $a ) )[0];
                    my $bkey = ( split( /=/, $b ) )[0];
                    return $akey cmp $bkey
                } @txtRecords
              )
            {

                # note, there CANNOT be an '=' in the key part...everything
                # up to the first '=' is the key; the rest is the value, which
                # may contain on or more '='
                if ( $txtRecord =~ /=/ ) {

                    # whittle down to just key=value (no newline nor surrounding quotes)
                    chomp $txtRecord;
                    $txtRecord =~ s/^"(.+)"$/$1/;

                    my ( $key, $value ) = split /=/, $txtRecord, 2;

                    # printMsg( 'DEBUG',"\nCTP KEY: $key VALUE: $value\n");
                    if (    exists $$results->{config}
                        and exists $$results->{config}->{$key}
                        and defined $$results->{config}->{$key} )
                    {

                        printMsg( "ERROR", "Got duplicate key for '$key'; will not continue" );
                        stunnel_completed($OPTS);
                        return 1;
                    }
                    $$results->{config}->{$key} = $value;
                    $sumTXT .= $txtRecord
                      unless $key eq 'checksum';    # don't include the checksum TXT
                }
            }

            # make sure not altered in any way
            # chomp(my $sha1 = `which sha1`);
            # &printMsg('DEBUG', "sha1 app: '$sha1'");
            # my $checksum_cmd = "echo -n '$sumTXT' | $sha1 | cut -d' ' -f1";
            # &printMsg('DEBUG', "checksum cmd: $checksum_cmd");
            # chomp( my $localChecksum = `$checksum_cmd` );
            my $localChecksum = generate_sha1_checksum($OPTS, $sumTXT);
            if ( !defined $$results->{config}->{checksum}
                || $localChecksum ne $$results->{config}->{checksum} )
            {

                printMsg( "DEBUG", "Checksums do not match : $localChecksum (locally) vs $results->{config}->{checksum} (from payload)" );
                printMsg( "ERROR", "Automatic configuration was corrupted during transit and cannot be used" );
                stunnel_completed($OPTS);
                return 1;
            }

            stunnel_completed($OPTS);
            &printMsg('OK', "Retrieved web automation settings");
            return 0; # success
        }
        else {

            printMsg( "DEBUG", "DNS server '$dns' did not satisfy the AXFR" );
            foreach my $responseLine ( split( /\n/, $cmdout ) ) {

                printMsg( "DEBUG", $responseLine );
            }
        }
    }

    stunnel_completed($OPTS);

    if ( $dig_completed eq 0 ) {
        printMsg( "ERROR", "Failed to connect to Web Automation services (ts-ctp.threatstop.com, TCP/5353)\n" );
    }
    else {
        printMsg( "ERROR", "Failed to retrieve Web Automation settings (Device ID not found)\n" );
        printMsg( "INFO", "Verify system time is in sync with upstream NTP servers" );
        printMsg( "INFO", "Retry in 5 minutes. " );
    }

    printMsg( "DEBUG", "No good answers for '$cfgZone'" );
    return 1;
}

1;
