#$Id$
#Package Maintainer : support@threatstop.com
#########################################################################
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# United States Patent No. 62539504 (Pending),
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
########################################################################

=head1 SYNOPSIS

 This module is meant to abstract the specifics of a particular on-device install.
 This module will change from device to device, but should provide the same interface
 with at least the following methods.
=cut

package ThreatStop::BaseDevice;
use strict;
use warnings;
use 5.010;    #enable perl 5.10 features.

use Exporter;
our $VERSION = 1.00;
our @ISA     = qw(Exporter);
our @EXPORT =
  qw(device_abortExec device_preflight_check device_update device_doPrompts device_doAdvancedPrompts device_has_required_settings device_preinstall device_install device_reconfigure device_postinstall device_uninstall device_upgrade device_backup device_restore device_load_cli_opts device_load_environment device_show_settings $DEVICE_DEFAULTS $DEVICE_USAGE @DEVICE_REQUIRED_APPS @DEVICE_REQUIRED_OPTS @DEVICE_OPTS @DEVICE_DONT_WRITE_KEYS $action $DEVICE_WEBAUTO_OPTS $DEVICE_WEBAUTO_ADVANCED_OPTS @DEVICE_RECONFIGURE_ONCHANGE @DEVICE_UPDATE_ONCHANGE $DEVICE_NAME);

use ThreatStop;

# dev
use Data::Dumper;

# default values if not defined in the template or conf file.
our $DEVICE_DEFAULTS        = {};      # hash of option => default value if not supplied by config or cli opts
our $DEVICE_USAGE           = {};      # hash of option => description of options function
our @DEVICE_OPTS            = qw();    # all device specific settings used for user confirmation of settings
our @DEVICE_REQUIRED_OPTS   = qw();    # device specific required settings
our @DEVICE_REQUIRED_APPS   = qw();    # device specific required system executables
our @DEVICE_DONT_WRITE_KEYS = qw();    # keys that shouldnt be written to threatstop.conf

###################################################################################
############################  CTP SETTINGS ########################################
###################################################################################

# if a main webauto opt is defined here but not found in the response, it will fail.
our $DEVICE_WEBAUTO_OPTS = {};

# advanced opts are not gauranteed to be there.
our $DEVICE_WEBAUTO_ADVANCED_OPTS = {};

our @DEVICE_RECONFIGURE_ONCHANGE = qw(tsprefix logopts logopts_direction);    # list of settings to trigger reconfigure when changed via ctp update
our @DEVICE_UPDATE_ONCHANGE      = qw(tsprefix logopts logopts_direction);    # list of settings to trigger update when changed via ctp update
our $DEVICE_NAME                 = 'ts-iptables';                             # this is the package name i.e. tsadmin show version outputs: $DEVICE_NAME Version: x.xx-xx

our $action;    # a variable shared between libs to convey action user is attempting to take. "add, remove, update, etc.."

=head3 DEVICE_VALIDATE HASH

Used to validate user input. We have 3 types of validation mechanisms
 1. - regex - matching
 2. - not_regex - inverse matching
 3. - code anonymous function must return 1 for success, 0 for fail.
 4. - coderef to function must return 1 for success, 0 for fail.

 Note: $DEVICE_VALIDATE hash is read and merged into main $VALIDATE hash early on.

 order of operation listed above, first existing & defined hash value gets run.

 block_list => {
        regex     => '',
        not_regex => '',
        code      => '',
        coderef   => '',
    },


=cut

our $DEVICE_VALIDATE = {

    # 'dns.threatstop.local',
    block_list => {
        regex     => '',
        not_regex => '',
        code      => '',
        coderef   => '',
    }
};

sub device_abortExec {
    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

sub device_usage {
    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

=head3 device_preflight_check

 Performs any necessary / possible checks required to ensure we can run on
 this system.

 Args :
       hashref ($OPTS) - general settings hash

 Returns :
   0 - success
   1 - fail
=cut

sub device_preflight_check {
    my $OPTS = shift;

    &printMsg( 'DEBUG', 'Device Preflight Check' );
    return 0;

}

=head3 device_doPrompts

Runs all the device specific prompts to collect required settings.

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_doPrompts {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_doAdvancedPrompts

Runs all the device specific advanced prompts to collect required settings.

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_doAdvancedPrompts {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_has_required_settings

verifies all device specific required settings are present and validated

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_has_required_settings {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_postinstall

to be run just before install
=cut

sub device_preinstall {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

=head3 device_install

performs actual installation:
- log settings
- log rotation
- cron tasks (log_upload, SRV url update settings, telemetry, CTP persistent changes )
- create uninstall.sh if necessary

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_install {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

=head3 device_postinstall

to be run just before unlock & exit
=cut

sub device_postinstall {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

=head3 device_uninstall

if uninstall.sh exists it runs that, else performs actual uninstallation:
- remove log settings
- remove log rotation
- remove cron tasks (log_upload, SRV url update settings, telemetry, CTP persistent changes )
- remove uninstall.sh if used

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_uninstall {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_update - optional

Do anything required for an update:
    - get policy data
    - apply configuration on device
    - telemetry

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_update {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_reconfigure - optional

do anything required when a ctp setting requires you to reconfigure your device:

- remove previous
    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_reconfigure {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_upgrade - optional

do anything required for an upgrade:

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_upgrade {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_backup - optional

create backup configs of existing system, named 'pre-threatstop'.

    Args :
        none

    Returns :
        0 - success
        1 - fail
=cut

sub device_backup {

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_restore - optional

restore backup configs if they exist ('pre-threatstop') else warn.

    Args :
        none

    Returns :
        0 - success
        1 - fail
=cut

sub device_restore {

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_load_cli_opts

Loading of device specific options passed via CLI as arguments.
General order of option priority (1 - highest):
    1 - cli
    2 - environment settings (dig location etc..)
    3 - SRV record update
    4 - config

    Args :
        hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_load_cli_opts {
    my $OPTS = shift;

    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");

    # my @device_opts = ('tsprefix:s', 'dig:s', 'ipset:s', 'iptables:s', 'logopts:s');
    # return \@device_opts;

}

=head3 device_load_environment

this method loads environment settings like: dig, nsupdate, curl locations
    Args:
       hashref ($OPTS) - general settings hash

    Returns:
       0 - success
       1 - fail
=cut

sub device_load_environment {
    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

=head3 device_show_settings

show device specific settings used in both user confirmation and general 'tsadmin show'
    Args:
       hashref ($OPTS) - general settings hash

    Returns:
       0 - success
       1 - fail
=cut

sub device_show_settings {
    my $sub_name = ( caller(0) )[3];
    die("ERROR: method '$sub_name' has not been written yet");
}

1;
