#$Id$
#Package Maintainer : support@threatstop.com
#########################################################################
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# United States Patent No. 62539504 (Pending),
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
########################################################################

=head1 SYNOPSIS

 This module is meant to abstract the specifics of a particular on-device install.
 This module will change from device to device, but should provide the same interface
 with at least the following methods.

 required methods defined in BaseDevice module
=cut

package ThreatStop::LocalDevice;
use strict;
use warnings;
use File::Basename;
use 5.010;      # enable perl 5.10 features.
use version;    # provided by 5.10 for comparing symantic versions

# use File::Basename;  # provides basename, dirname & fileparse
use POSIX qw(strftime);

use Exporter;
our $DEVICE_VERSION = "2.00";

our @ISA = qw(Exporter);
our @EXPORT =
  qw(is_already_configured checkValidIP device_abortExec device_update device_reconfigure device_preflight_check device_doPrompts device_doAdvancedPrompts device_has_required_settings device_preinstall device_install device_postinstall device_uninstall device_upgrade device_backup device_restore device_load_cli_opts device_load_environment device_show_settings $DEVICE_DEFAULTS $DEVICE_VALIDATE @DEVICE_REQUIRED_APPS @DEVICE_REQUIRED_OPTS @DEVICE_OPTS @DEVICE_DONT_WRITE_KEYS $DEVICE_VERSION $SERVICE_ACCOUNT $SERVICE_GROUP $SU_ACCOUNT $action $DEVICE_WEBAUTO_OPTS $DEVICE_WEBAUTO_ADVANCED_OPTS $DEVICE_USAGE @DEVICE_UPDATE_ONCHANGE @DEVICE_RECONFIGURE_ONCHANGE $DEVICE_NAME $DEVICE_TYPE);
use Getopt::Long;    # requires libgetopt-simple-perl

use ThreatStop;

# use ThreatStop::BaseDevice;

# dev
use Data::Dumper;

no warnings 'redefine';    # this suppresses warning about inherited BaseDevice subs we redefine here

#TODO: do we need to provide $LOG_DIR $LOG_BASENAME ?

our $DEVICE_TYPE = 'pfsense';                                            # this is the package name i.e. tsadmin show version outputs: $DEVICE_NAME Version: x.xx-xx
our $DEVICE_NAME = 'ts-' . $DEVICE_TYPE;                                        # this is the package name i.e. tsadmin show version outputs: $DEVICE_NAME Version: x.xx-xx

# default values if not defined in the template or conf file.
our $DEVICE_DEFAULTS = {
    logfile => '/var/log/filter.log',
    devtype => $DEVICE_TYPE,
};

our $DEVICE_USAGE = {
    nsupdate          => 'specify an absolute disk path to executable system application',
    dig               => 'specify an absolute disk path to executable system application',
    skip_checkvalidip => 'Skip validating your public IP is registered with ThreatSTOP',
};

our @DEVICE_OPTS            = qw();    # all device specific settings used for user confirmation of settings
our @DEVICE_REQUIRED_OPTS   = qw(logfile);                               # device specific required settings
our @DEVICE_REQUIRED_APPS   = qw(perl php);                   # device specific required system executables
our @DEVICE_DONT_WRITE_KEYS = qw();                                      # keys that shouldnt be written to threatstop.conf


###################################################################################
############################  CTP SETTINGS ########################################
###################################################################################

# if a main webauto opt is defined here but not found in the response, it will fail.
our $DEVICE_WEBAUTO_OPTS = { devtype => 'devtype' };

# advanced opts are not gauranteed to be there.
our $DEVICE_WEBAUTO_ADVANCED_OPTS = {
    port => 'port',
};

our @DEVICE_RECONFIGURE_ONCHANGE = qw(logupload telemetry);    # list of settings to trigger reconfigure when changed via ctp update
our @DEVICE_UPDATE_ONCHANGE      = qw(block_list allow_list);    # list of settings to trigger update when changed via ctp update

our $action;    # a variable shared between libs to convey action user is attempting to take. "add, remove, update, etc.."

=head3 DEVICE_VALIDATE HASH

Used to validate user input. We have 3 types of validation mechanisms
 1. - regex - matching
 2. - not_regex - inverse matching
 3. - code anonymous function must return 1 for success, 0 for fail.
 4. - coderef to function must return 1 for success, 0 for fail.

 Note: $DEVICE_VALIDATE hash is read and merged into main $VALIDATE hash early on.

 order of operation listed above, first existing & defined hash value gets run.

 block_list => {
        regex     => '',
        not_regex => '',
        code      => '',
        coderef   => '',
    },


=cut

our $DEVICE_VALIDATE = {
    # CTP settings are for this type of device
    devtype => {
        regex     => '^pfsense$',
        not_regex => '',
        code      => '',
        coderef => '',
    },
};

our $SERVICE_ACCOUNT = 'threatstop';
our $SERVICE_GROUP = 'threatstop';
my $SU_ACCOUNT      = 'root';

=head3 device_abortExec

do what is necessary to safetly abort and exit mid-setup

=cut

sub device_abortExec {
    my ( $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    &printMsg('START', "device_abortExec");
    if (defined $action and $action eq 'add') {
        &printMsg('DEBUG', "abort - removing any installed configs");
        &printMsg( 'FAIL',  "removing scheduler" ) if ( uninstall_scheduler($OPTS) != 0 );
    }
}

# sub validate_policy {
#     my ($OPTS) = shift;
#     my $cmd = "$OPTS->{'dig'} -p $OPTS->{'port'} ptr \@$OPTS->{dns_server} $OPTS->{block_list}";
#     &printMsg( "DEBUG", "$cmd" );
#     my $res = qx{$cmd 2>&1};
#     if ( $res !~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s ) {
#         &printMsg( "FAIL", "Testing dig access to $OPTS->{block_list}" );
#         return 1;
#     }
#     else {
#         &printMsg( "OK", "Testing dig access to $OPTS->{block_list}" );
#         return 0;
#     }
# }

=head3 device_preflight_check

 Performs any necessary / possible checks required to ensure we can run on
 this system.

 Args :
       hashref ($OPTS) - general settings hash

 Returns :
   0 - success
   1 - fail
=cut

sub device_preflight_check {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    my @required_apps = @DEVICE_REQUIRED_APPS;
    # for my $app (@required_apps) {
    for my $app (@DEVICE_REQUIRED_APPS) {
        &printMsg( 'CHECK', "detecting $app" );
        if ( check_for_app( $app, 'is_fatal', $OPTS ) == 0 ) {
            my $app_version = get_app_version( $OPTS->{$app} );
            &printMsg( 'CHECK', "$app version [$app_version]" );
            $OPTS->{"${app}_version"} = $app_version if $app_version;
        }
    }

    my @desired_apps = qw(pfctl crontab clog shasum);
    for my $app (@desired_apps) {
        &printMsg( 'CHECK', "detecting '$app'" );
        if ( check_for_app( $app, undef, $OPTS ) == 0 ) {

            # get version for all of these.
            next if $app eq 'crontab';    # can't get crontab's version
            my $app_version = get_app_version( $OPTS->{$app} );
            &printMsg( 'CHECK', "$app version [$app_version]" );
            $OPTS->{"${app}_version"} = $app_version if $app_version;
        }
    }

    my @menu_hooks = ("/usr/local/www/fbegin.inc", "/usr/local/www/head.inc");
    my $found_hook = 0;
    for my $hook (@{menu_hooks}) {
        if (-f "$hook") {
            &printMsg('DEBUG', "Detected menu hook: $hook");
            $OPTS->{'menu_hook'} = $hook;
            $found_hook = 1;
        }
    }

    &printMsg('DEBUG', "No menu hook file found. " . join(' ', @menu_hooks)) unless $found_hook;

    abortExec() unless $found_hook;

    if (! exists $OPTS->{shasum}) {
        &printMsg('ERROR', "shasum for CTP crc checks");
        abortExec();
    }

    # update our local device version if $TSDIR/VERSION and $TSDIR/PATCH exist
    if (exists $OPTS->{src} and $OPTS->{src}) {
        if (-f "$OPTS->{src}/VERSION" and -f "$OPTS->{src}/PATCH") {
            chomp(my $local_version = `cat $OPTS->{src}/VERSION`);
            chomp(my $local_revision = `cat $OPTS->{src}/PATCH`);
            $DEVICE_VERSION = "${local_version}.${local_revision}";
        }
    }

    # check we are running pfsense 2.4 or other supported versions only
    chomp(my $pfsense_version = `cat /etc/version`);
    if ($pfsense_version =~ m/^2.4/) {
        &printMsg('OK', "pfSense version [$pfsense_version]");
    } else {
        &printMsg('ERROR', "pfSense version [$pfsense_version] not supported");
        abortExec();
    }

    return 0;
}

=head3 device_doPrompts

Runs all the device specific prompts to collect required settings.

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_doPrompts {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    # print_banner('Device Settings', 'cyan');
    # $OPTS->{'tsprefix'} = input_prompt( 'Please enter the ThreatSTOP iptables / ipset name prefix', '^[\w\-\_\.]+$', $OPTS->{tsprefix} || $DEVICE_DEFAULTS->{tsprefix} );

}

=head3 device_doAdvancedPrompts

Runs all the device specific advanced prompts to collect required settings.

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_doAdvancedPrompts {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    # print_colorized( ' ' x 32 . "Device Advanced Settings" . ' ' x 32 . "\n", 'cyan', 'reverse' );

    # $OPTS->{'logopts'} = input_prompt( 'iptables logging options:', '^\-\-log\-[\w\-\_]+|^$', $OPTS->{logopts} || $DEVICE_DEFAULTS->{logopts}, 0, 1 );
    # if ( exists $OPTS->{'logopts'} and $OPTS->{'logopts'} ) {

    #     # prompt for inbound, outbound or both
    #     $OPTS->{'logopts_direction'} = input_prompt( 'iptables logging options direction valid choices (in, out, both):',
    #         '^in$|^out$|^both$', $OPTS->{logopts_direction} || $DEVICE_DEFAULTS->{logopts_direction} );
    # }

    # my $sub_name = ( caller(0) )[3];
    # die("ERROR: method '$sub_name' has not been written yet");

}

=head3 device_has_required_settings

verifies all device specific required settings are present and validated

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_has_required_settings {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    # nothing special to do for this build... return OK - 0

    return 0;

}

=head3 device_postinstall

to be run just before install
=cut

sub device_preinstall {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    # Check all the CAN we installs
    # safety, backup existing iptables & ipset configurations first
    device_backup($OPTS);

    # Check validIP test should be done here since the lookup tool may vary per device
    if ( $OPTS->{'skip_checkvalidip'} ) {
        &printMsg( "INFO", "Bypassing public IP access check." );
    }
    else {
        unless ( checkValidIP($OPTS) ) {
            my $troubleshooting =
"\n\tTroubleshooting:\n\t1. Make sure you have a valid ThreatSTOP account.\n\t2. Make sure you have created a device entry on the ThreatSTOP Admin Portal (http://admin.threatstop.com).\n\t3. Attempt to run 'curl $OPTS->{validip_url}' and verify that IP matches the device public IP set on the ThreatSTOP admin portal device entry.\n\t4. Verify at least 15minutes has elapsed since adding device to admin portal.\n\n";
            &printMsg( "ERROR", "This device's public IP can not access our servers. $troubleshooting" );
            &printMsg( "INFO",  "To bypass this check add the --skip_checkvalidip flag." );
            exit 2;
        }
        else {
            &printMsg( "OK", "Public IP can access our servers." );
        }
    }

    check_DNS_access($OPTS);

    # Check all the SHOULD we installs
    # check if we've already installed service on this device
    $OPTS->{'is_already_configured'} = is_already_configured($OPTS);
    if ( $OPTS->{'is_already_configured'} ) {
        &printMsg( 'ERROR', "already configured" );
        print_colorized( "ThreatSTOP must be uninstalled before ${action}ing is performed.\n\tRun 'tsadmin remove' to uninstall.\n", 'red' );
        return 1;
    }

    return 0;
}

=head3 device_install

performs actual installation:
- log settings
- log rotation
- cron tasks (log_upload, SRV url update settings, telemetry, CTP persistent changes )
- create uninstall.sh if necessary

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_install {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    &abortExec() if ( install_pfsense($OPTS) != 0 );
    &abortExec() if ( install_scheduler($OPTS) != 0 );
    # &printMsg('DEBUG', "TINA, EAT THE FOOD!!" . Dumper $OPTS); die('****** STOP HERE FOR DEV TESTING *******');

}

=head3 device_postinstall

to be run just before unlock & exit
=cut

sub device_postinstall {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    &printMsg( 'DEBUG', "setting is_already_configured flag" );
    $OPTS->{is_already_configured} = 1;
    $OPTS->{service_account}       = $SERVICE_ACCOUNT;

    return 0;
}

=head3 device_uninstall

if uninstall.sh exists it runs that, else performs actual uninstallation:
- remove log settings
- remove log rotation
- remove cron tasks (log_upload, SRV url update settings, telemetry, CTP persistent changes )
- remove uninstall.sh if used

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_uninstall {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    if ($OPTS->{auto_remove}) {
        &printMsg('DEBUG', "auto-removing..");
        uninstall_pfsense($OPTS);
        &printMsg( 'FAIL',  "removing scheduler" ) if ( uninstall_scheduler($OPTS) != 0 );
    } else {
        print_banner('WARNING', 'bold_red');
        print_colorized("\n\tIf you have not done so already, log into the pfSense-web interface\n\tand click 'Disable', then 'Remove' before running this command.\n\n", 'red');
        if ( confirm_prompt("Continue removing ThreatSTOP Integration files?", 'n') ) {
            &printMsg('INFO', "Removing Integration files...");
            &printMsg('DEBUG', "user chose to continue uninstall");
            uninstall_pfsense($OPTS);
            &printMsg( 'FAIL',  "removing scheduler" ) if ( uninstall_scheduler($OPTS) != 0 );
        }
        else {
            &printMsg('DEBUG', "user chose to abort uninstall");
            &setUnLock();
            print_colorized( "\n\tUninstallation aborted. Installation untouched...\n", 'red' );
            exit 0;
        }

    }



    return 0;
}

=head3 device_update - optional

Do anything required for an update:
    - get policy data
    - apply configuration on device
    - telemetry

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_update {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    return update_policy($OPTS);

    return 0;
}

=head3 device_upgrade - optional

do anything required for an upgrade:

    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_upgrade {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    &printMsg('DEBUG', "no upgrade method for $DEVICE_NAME");

    return 0;
}

=head3 device_reconfigure - optional

do anything required when a ctp setting requires you to reconfigure your device:

# note: the OPTS hash will contain keys starting with reconfigure_ that require them to remain the same until uninstalled & reinstalled.

- remove previous
    Args :
       hashref ($OPTS) - general settings hash

    Returns :
        0 - success
        1 - fail
=cut

sub device_reconfigure {
    my ( $self, $OPTS ) = @_;

    my $rval     = 0;
    my $sub_name = ( caller(0) )[3];
    &printMsg( 'START', "Running $sub_name" );

    # load up reconfigure keys with their original key name.
    my $reconfig;
    map { my $key = $_ if $_ =~ m/^reconfigure_/; my $orig_key = $key; $orig_key =~ s/reconfigure_//g if $key; $reconfig->{$orig_key} = delete $OPTS->{$key} if $key; }
      keys( %{$OPTS} );

    unless ( keys %{$reconfig} ) {
        &printMsg( 'DEBUG', "No changes detected. Exiting now." );
        return 0;
    }

    my $cmd = "$OPTS->{'dig'} -p $OPTS->{'port'} ptr \@$OPTS->{dns_server} $OPTS->{block_list}";
    my $res = qx{$cmd 2>&1};
    &printMsg( "DEBUG", "$cmd" );

    # Check for possible DNS query call errors
    if ( $res !~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s ) {
        set_f_telemetry( 'policy_download_status', 'failure' );
        &printMsg( "FAIL", "Testing dig access to $OPTS->{block_list}" );
        return 1;
    }
    else {
        &printMsg( "OK", "Testing dig access to $OPTS->{block_list}" );
    }
    # if ( uninstall_pfsense($OPTS) != 0 ) {
    #     $rval = 1;
    #     &printMsg( 'FAIL', ">> Reconfiguring" );
    # }

    &printMsg( 'DEBUG', "Writing staged keys to config." );
    map {
        &printMsg( 'DEBUG', "Committing '$_' => '$reconfig->{$_}' " );
        $OPTS->{$_} = $reconfig->{$_};
        write_conf_key( "$OPTS->{src}/threatstop.conf", { $_ => $reconfig->{$_} } )
    } keys( %{$reconfig} );

    # new ipset settings will be installed with update
    if (install_scheduler($OPTS) != 0 )  {
        &printMsg( 'FAIL',  "Installing scheduler" );
        return 1;
    }
    else {
            &printMsg( 'OK',  "Installing scheduler" );
    }
    # if ( install_pfsense($OPTS) == 0 ) {
    #     &printMsg( 'OK', ">> Reconfiguring" );
    #     system("printf 'reconfigured $OPTS->{devname}' |wall");
    # }
    # else {
    #     $rval = 1;
    #     &printMsg( 'FAIL', ">> Reconfiguring: installing rules" );
    # }
    system("$SUDO printf 'ThreatSTOP Web Auto reconfiguration detected $OPTS->{devname}' |wall");

    return $rval;

}

=head3 device_backup - optional

create backup configs of existing system, named 'pre-threatstop'.

    Args :
        none

    Returns :
        0 - success
        1 - fail
=cut

sub device_backup {
    my ($OPTS) = @_;

    # backup iptables and ipset configurations
    my $errs            = 0;
    # TODO: tar up /etc/inc and /usr/local/www
    my $inc_backup    = "$OPTS->{src}/pre-threatstop-inc.tgz";
    my $www_backup = "$OPTS->{src}/pre-threatstop-www.tgz";

    my $tar = $OPTS->{'tar'} ? $OPTS->{'tar'} : `which tar`;
    chomp($tar);

    if ( !-f $www_backup ) {
        my $www_cmd = "$SUDO $tar -czvf $www_backup /usr/local/www > /dev/null 2>&1";
        &printMsg( "DEBUG", "device_backup running cmd: $www_cmd" );
        system("$www_cmd") && print "$!\n";
        my $exit_code = $?;
        $errs += $exit_code if $exit_code;    # if we have anything other than clean exit 0 this will add it to return status

        &printMsg( "OK", "backup: $www_backup created" ) if $exit_code == 0;
    }
    else {
        &printMsg( "DEBUG", "backup: $www_backup already exists" );
    }

    if ( !-f $inc_backup ) {
        my $inc_cmd = "$SUDO $tar -czvf $inc_backup /etc/inc > /dev/null 2>&1";
        &printMsg( "DEBUG", "device_backup running cmd: $inc_cmd" );
        system("$inc_cmd") && print "$!\n";
        my $exit_code2 = $?;
        $errs += $exit_code2 if $exit_code2;    # if we have anything other than clean exit 0 this will add it to return status

        &printMsg( "OK", "backup: $inc_backup created" ) if $exit_code2 == 0;
    }
    else {
        &printMsg( "DEBUG", "backup: $inc_backup already exists" );
    }

    return $errs;

}

=head3 device_restore - optional

restore backup configs if they exist ('pre-threatstop') else warn.
NOTE: ipset groups MUST be flush'd and destroy'd prior to this step. If we detect otherwise we abort.

    Args :
        none

    Returns :
        0 - success
        1 - fail
=cut

sub device_restore {

    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    # backup iptables and ipset configurations
    my $errs            = 0;
    my $inc_backup    = "$OPTS->{src}/pre-threatstop-inc.tgz";
    my $www_backup = "$OPTS->{src}/pre-threatstop-www.tgz";

    my $tar = $OPTS->{'tar'} ? $OPTS->{'tar'} : `which tar`;
    chomp($tar);

    unless ( -f $inc_backup ) {
        &printMsg( 'ERROR', "No backup found: $inc_backup" );
        return 1;
    }
    if ( $OPTS->{is_already_configured} and is_already_configured($OPTS) ) {
        &printMsg( 'ERROR', "ThreatSTOP must be uninstalled first" );
        print_colorized( "ThreatSTOP must be uninstalled before ${action}ing is performed.\n\tRun 'tsadmin remove' to uninstall.\n", 'red' ) if $OPTS->{is_already_configured};
        &abortExec();
    }
    else {
        if ( confirm_prompt('Do you want to restore configuration backups taken prior to initial installation?') ) {
            &printMsg( 'DEBUG', "User chose to restore $www_backup & $inc_backup" );
        }
        else {
            &printMsg( 'INFO', "Nothing done." );
            return 0;
        }
    }

    # clear to restore

    if ( -f $inc_backup ) {
        my $inc_cmd = "cd / && $SUDO $tar -xzvf $inc_backup";
        &printMsg( "DEBUG", "device_restore running cmd: $inc_cmd" );
        system("$inc_cmd") && print "$!\n";
        my $inc_exit = $?;
        $errs += $inc_exit if $inc_exit;    # if we have anything other than clean exit 0 this will add it to return status

        &printMsg( "OK", "restore: $inc_backup restored" ) if $inc_exit == 0;
    }
    else {
        &printMsg( "WARN", "restore: $inc_backup did not exists. Not restored." );
    }

    if ( -f $www_backup ) {
        my $www_cmd = "cd / && $SUDO $tar -xzvf $www_backup";
        &printMsg( "DEBUG", "device_restore running cmd: $www_cmd" );
        system("$www_cmd") && print "$!\n";
        my $www_exit = $?;
        $errs += $www_exit if $www_exit;    # if we have anything other than clean exit 0 this will add it to return status

        &printMsg( "OK", "restore: $www_backup restored" ) if $www_exit == 0;
    }
    else {
        &printMsg( "WARN", "restore: $www_backup did not exists. Not restored." );
    }

    my $pfsense_config = "/cf/conf/config.xml";
    if (-f $pfsense_config) {
        restore_file($pfsense_config);
    }

    return $errs;
}

=head3 device_load_cli_opts

Loading of device specific options passed via CLI as arguments.
General order of option priority (1 - highest):
    1 - cli
    2 - environment settings (dig location etc..)
    3 - SRV record update
    4 - config

    Args :
        hashref ($OPTS) - general settings hash
    Returns :
        array ref - list of options to load (will be merged into global options)
=cut

sub device_load_cli_opts {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");


    my @device_opts = ( 'skip_checkvalidip' );
    return \@device_opts;
}

=head3 device_load_environment

this method loads environment settings like: dig, nsupdate, curl locations
    Args:
       hashref ($OPTS) - general settings hash

    Returns:
       0 - success
       1 - fail
=cut

sub device_load_environment {
    my ( $self, $OPTS ) = @_;

    my $sub_name = ( caller(0) )[3];
    &printMsg('START', " [Running - $sub_name]");

    my @get_app_version = qw(curl dig host);    # only apps we may need version for
    my @desired_apps =
      qw(tar install ntpdate stunnel stunnel4 nsupdate curl dig host service cpan cpanm at kill clear);
    for my $app (@desired_apps) {
        if ( check_for_app( $app, undef, $OPTS ) == 0 ) {
            if ( grep { $_ eq $app } @get_app_version ) {
                my $app_version = get_app_version( $OPTS->{$app} );
                &printMsg( 'DEBUG', "$app version [$app_version]" );
                $OPTS->{"${app}_version"} = $app_version if $app_version;
            }
        }
    }

    # # finds rsyslog, syslog-ng
    # my @supported_loggers = qw(rsyslog syslog-ng);
    # detect_app_info( $OPTS, 'logger',  @supported_loggers );
    detect_app_info( $OPTS, 'rotater', qw(logrotate) );
    $OPTS->{'logger_restart_cmd'} = get_app_restart( $OPTS, $OPTS->{logger} ) if exists $OPTS->{logger};

    # # $OPTS->{'rotater_restart_cmd'} = get_app_restart($OPTS, 'logrotate') if exists $OPTS->{logrotate}; # no need to restart logrotate its usually called via cron
    # # finds crontab
    detect_app_info( $OPTS, 'scheduler', qw(crontab) );

    return 0;

}

=head3 device_show_settings

show device specific settings used in both user confirmation and general 'tsadmin show'
    Args:
       hashref ($OPTS) - general settings hash

    Returns:
       0 - success
       1 - fail
=cut

sub device_show_settings {
    my ( $self, $OPTS ) = @_;
    &printMsg( "DISPL", "Logfile location       : $OPTS->{'logfile'}" );

    # my $mode = {
    #     r => 'router',
    #     b => 'bridge',
    #     u => 'unchanged'
    # };
    # &printMsg( "DISPL", "Operating Mode         : $mode->{$OPTS->{'mode'}}" );
    # &printMsg( "DISPL", "Default allow action   : $OPTS->{'allow_action'}" );
    # &printMsg( "DISPL", "Log options (iptables) : $OPTS->{'logopts'}" );
    # my $logopts_direction = (exists $OPTS->{logopts} and $OPTS->{logopts} ) ? $OPTS->{'logopts_direction'} : '';
    # &printMsg( "DISPL", "Log opts direction     : $logopts_direction" );
    # &printMsg( "DISPL", "Prefix (iptables)      : $OPTS->{'tsprefix'}" );
    # &printMsg( "DISPL", "Load bridge module     : $OPTS->{'load_br_netfilter'}" );
    # &printMsg( "DISPL", "ipset version          : $OPTS->{'ipset_version'}" );
    # &printMsg( "DISPL", "iptables version       : $OPTS->{'iptables_version'}" );
    &printMsg( "DISPL", "dig version            : $OPTS->{'dig_version'}" );
    &printMsg( "DISPL", "perl version           : $OPTS->{'perl_version'}" );

}

############################################################################
##################  HELPER METHODS FOR SUBS ABOVE ##########################
############################################################################

# MAKE sure you don't export abortExec, it should be overridden by local scripts version
sub abortExec() {

    &setUnLock();
    print_colorized( "\n\tInstallation aborted...\n", 'red' );
    exit 1;
}

=head3 checkValidIP

    checkValidIP - curls our valid IP endpoint to see if customers public IP is whitelisted.

    args:
        none - but does rely on $OPTS (configuration to be read into memory)
    returns:
        0 - failure - can NOT access our servers
        1 - success - can access our servers

=cut

sub checkValidIP {
    my $OPTS = shift;
    &printMsg( 'CHECK', "Validating Public IP can access our servers");
    my $validipurl = $OPTS->{'validip_url'} || 'https://logs.threatstop.com/validip.pl';
    &printMsg( 'DEBUG', "Running: $OPTS->{'curl'} -s $validipurl" );
    chomp(my $response = `$OPTS->{'curl'} -s $validipurl|head -2|tail -1`);
    &printMsg('DEBUG', "validip response: '$response'");

    # reponse is either 'Address is in the list of authorized hosts' or 'Address is not in the list of authorized hosts'
    my $has_access = ( defined $response and $response and $response =~ m/Address is in the list/ ) ? 1 : 0;
    &printMsg('DEBUG', "has_access: $has_access");
    return $has_access;
}

#TODO: Fix this including $DNS_LIST

=head4 check_DNS_access

testingConnection : test access to DNS servers    #
      $DNS_LIST is default variable loaded from   #
                lib/ThreatStop.pm module          #
=cut

sub check_DNS_access {
    my $OPTS = shift;

    my @dns_ok_list;

    #Testing DNS servers access. Also verify dig command
    #loaded from lib/Threatstop.pm module
    if ( &checkDNSaccess( $OPTS->{'dig'}, $OPTS->{'dns_server'}, \@dns_ok_list, $OPTS->{'port'} ) != 0 ) {
        &printMsg( 'OK', 'DNS Access via dig' );
        return 1;
    }

    &printMsg( "DEBUG", "Using DNS servers list [@dns_ok_list]" );
    return 0;
}

=head3 is_already_configured

is_already_configured just checks if we have allownet ipset rules.
TODO: check for config files and iptables rules

    args:
        none
    returns:
        1 - is configured / true
        0 - is NOT configured / false
=cut

sub is_already_configured {

    # my ($self, $OPTS) = @_;
    my $OPTS          = shift;
    my $is_configured = 0;
    my @installable_files = ('tslogupload.php', 'tsgetblockip.php', 'tsenable.php', 'ts_state.php', 'threatstop.conf.php');
    my $install_base_dir = '/usr/local/www';
    for my $tsfile (@{installable_files}) {

        if (-f "$install_base_dir/$tsfile") {
            &printMsg('DEBUG', "Found $tsfile file installed.");
            $is_configured = 1;
            last;
        }
    }
    return $is_configured;
}

sub install_scheduler {
    my ($OPTS) = @_;

    my $now_minute = strftime "%M", localtime;
    my $near_future = $now_minute;
    # we want logupload to be 15min in future from time of install
    if ( $near_future > 45 ) {
        $near_future = ($near_future - 45) + 15;
    }
    else {
        $near_future += 15;
    }
    if (($near_future % 5) == 0) {
        $near_future += 2;
    } # we want non-interval of 5 so auto_config doesn't collide
    my $every_five    = '*/5';
    my $every_fifteen = '*/15';

    my $template       = "$OPTS->{src}/templates/crontab/threatstop_" . $DEVICE_TYPE;
    my $is_ctp_install = is_ctp_install($OPTS);

    my @config;
    push @config, "\\\@reboot     $SERVICE_ACCOUNT $OPTS->{perl} $OPTS->{'src'}/ts-update-policy\n\n";
    push @config, "# Update the ThreatSTOP block & allow list.  Run every hour\n";
    push @config, "$now_minute     *     *     *     *     $SERVICE_ACCOUNT $OPTS->{perl} $OPTS->{'src'}/ts-update-policy >> $TSLOGDIR/update_policy.log 2>&1 && tail -n 10000 $TSLOGDIR/update_policy.log > $TSLOGDIR/update_policy.log.tmp && mv $TSLOGDIR/update_policy.log.tmp $TSLOGDIR/update_policy.log\n\n";
    push @config, "# log upload runs spools every 5 uploads every 15 minutes\n";
    push @config, "$every_five  *     *     *     *    $SERVICE_ACCOUNT $OPTS->{'php'} /usr/local/www/tslogupload.php |tee -a $TSLOGDIR/logupload.log 2>&1 && tail -n 10000 $TSLOGDIR/logupload.log > $TSLOGDIR/logupload.log.tmp && mv $TSLOGDIR/logupload.log.tmp $TSLOGDIR/logupload.log\n\n"
      if exists $OPTS->{'logupload'} and $OPTS->{'logupload'} eq 'enabled';
    push @config, "# Update the config file with values from SRV query every 6 hours\n";
    push @config, "$near_future    */6     *     *     *     $SERVICE_ACCOUNT $OPTS->{perl} $OPTS->{'src'}/ts-update-settings >> $TSLOGDIR/update_settings.log 2>&1 && tail -n 10000 $TSLOGDIR/update_settings.log > $TSLOGDIR/update_settings.log.tmp && mv $TSLOGDIR/update_settings.log.tmp $TSLOGDIR/update_settings.log\n";
    push @config, "# Update the config file with values from Web Automation server every 5 min\n" if $is_ctp_install;
    push @config, "$every_five    *     *     *     *     $SERVICE_ACCOUNT $OPTS->{perl} $OPTS->{'src'}/ts-update-settings auto_config >> $TSLOGDIR/update_auto_settings.log 2>&1 && tail -n 10000 $TSLOGDIR/update_auto_settings.log > $TSLOGDIR/update_auto_settings.log.tmp && mv $TSLOGDIR/update_auto_settings.log.tmp $TSLOGDIR/update_auto_settings.log\n" if $is_ctp_install;
    &printMsg( 'DEBUG', "telem: $OPTS->{'telemetry'}" );
    &printMsg( 'DEBUG', "is_ctp: $is_ctp_install" );

    if ( "$OPTS->{telemetry}" eq "enabled" and $is_ctp_install ) {
        &printMsg( 'DEBUG', "INSTALLING TELEMETRY *** " );
        push @config, "# Hourly Update telemetry data\n";
        push @config, "$near_future  *     *     *     *     $SERVICE_ACCOUNT $OPTS->{perl} $OPTS->{'src'}/ts-telemetry |tee -a $TSLOGDIR/telemetry.log 2>&1 && tail -n 10000 $TSLOGDIR/telemetry.log > $TSLOGDIR/telemetry.log.tmp && mv $TSLOGDIR/telemetry.log.tmp $TSLOGDIR/telemetry.log\n";
    }
    my $config = join( "\n", @config );

    my $search_and_replace = { __TS_CRONTAB__ => $config };

    return install_app_config( $OPTS, 'scheduler', $template, $search_and_replace );
}

sub uninstall_scheduler {
    my ($OPTS) = @_;
    return uninstall_app_config( $OPTS, 'scheduler' );
}


sub update_policy {
    my $OPTS = shift;

    my @install_cmds;
    # push @install_cmds, ['running tssetup', "cd /opt/threatstop/templates/pfsense && $SUDO $OPTS->{'php'} ./tssetup.php"];
    my $update_cmd;
    $update_cmd = "$OPTS->{src}/ts-update-policy";
    $update_cmd .= " DEBUG" if ( $DEBUG_LEVEL > 5 );
    $update_cmd .= " SILENT" if ( $OPTS->{silent} );

    push @install_cmds, ['fetching allow / block lists', $update_cmd];

    for my $dscript_cmd (@{install_cmds}) {

        my $msg = $dscript_cmd->[0];
        my $cmd = $dscript_cmd->[1];

        my $rv;

        &printMsg( "INFO", "Starting $msg." );
        &printMsg( 'DEBUG', "install_pfsense running: $cmd" );

        if ( system("$cmd 2>&1") == 0 ) {
            &printMsg( 'OK', "$cmd" );
        }
        else {
            &printMsg( 'FAIL', "$cmd");
            &printMsg('DEBUG', "This may be normal if the device was recently added or the policy recently assigned" );
            &printMsg('INFO', "It may take up to 15 minutes to propogate a new policy" );
            return 1;
        }

        &printMsg( "INFO", "Finished $msg." );
    }
    return 0;

}


#####################################################################
#getLists: running ts-update-policy script, getting allow/block lists    #
#####################################################################
sub install_pfsense {
    my $OPTS = shift;

    # copy pfsense templates into /usr/local/www
    # merge into menu hook file (head.inc, or fbegin.inc)
    my $chown = 'root:wheel';
    my $chmod = 644;
    my $template_dir = "$OPTS->{src}/templates/pfsense";
    my $target_dir = '/usr/local/www';
    my $threatstop_logo = "${template_dir}/ts_logo_bw.png";
    if ( -f $threatstop_logo and ! -f "$target_dir/ts_logo_bw.png") {
        &printMsg('DEBUG', "Copying threatstop logo");
        system("$SUDO cp $threatstop_logo $target_dir && $SUDO chmod $chmod $target_dir/ts_logo_bw.png && $SUDO chown $chown $target_dir/ts_logo_bw.png");
    } else {
        &printMsg('DEBUG', "logo already in place.");
    }
    my $threatstop_menu_link = "firewall_menu[] = array(\\\"ThreatSTOP \\\", \\\"/ts_state.php\\\");\\\n\\\$firewall_menu = msort";
    my $search_and_replace = {'%clog%' => $OPTS->{clog}, 'firewall_menu = msort' => $threatstop_menu_link };
    my $source_to_target_conf = {};

    # make backup of xml just in case
    my $pfsense_config_backup = "$OPTS->{'src'}/pre-threatstop-config.xml";
    my $pfsense_config = "/cf/conf/config.xml";
    if (! -f "$pfsense_config_backup" and -f $pfsense_config ) {
        if (system("cp $pfsense_config $pfsense_config_backup") == 0) {
            &printMsg('OK', "backup of config to $pfsense_config_backup");
            backup_file($pfsense_config);
        } else {
            &printMsg('FAIL', "backup of config to $pfsense_config_backup");
        }
    } else {
        &printMsg('DEBUG', "backup $pfsense_config_backup already exists. skipping.");
    }

    # stage menu hook changes
    my $hook; # menu hook file
    if (-f "$target_dir/head.inc") {
        $hook = 'head.inc';
    } elsif (-f "$target_dir/fbegin.inc") {
        $hook = 'fbegin.inc';
    }
    my $hook_path = "$target_dir/$hook";
    backup_file($hook_path);
    system("$SUDO cp $hook_path $hook_path.old"); # create .old backup
    system("$SUDO cp $hook_path $template_dir/$hook"); # create .old backup

    #gui_cfg mods to change mouse-over popup ACL to show only 100 ips
    my $gui_cfg;
    if (-f "$target_dir/guiconfig.inc") {
        $gui_cfg = "$target_dir/guiconfig.inc";
        backup_file($gui_cfg);
        &printMsg('DEBUG', "setting gui popup to 100 ips only");
        system("$SUDO perl -i -pe 's/10k/100/g;' -pe 's/10002/102/g;' $gui_cfg"); # create .old backup
    }

    $source_to_target_conf->{"$template_dir/$hook"} = $hook_path;

    # stage filter.inc changes
    my $filter_path = '/etc/inc/filter.inc';
    if (-f $filter_path) {
        backup_file($filter_path);
        $search_and_replace->{"-ttt -i pflog0 \\| logger -t pf -p local0.info"} = '-ttt -i pflog0 \\| /usr/bin/sed -e "N;s/\\\\n //;P;D;" \\| logger -t pf -p local0.info';
        # $search_and_replace->{"^ = \"{  }\""} = '';
        system("$SUDO cp $filter_path ${template_dir}/"); # create .old backup
        $source_to_target_conf->{"$template_dir/filter.inc"} = $filter_path;
    }

    my @installable_files = ('bootstrap-toggle.min.css', 'bootstrap-toggle.min.js', 'tsremove.php', 'tslogupload.php', 'tsgetblockip.php', 'tsenable.php', 'tsdisable.php', 'ts_state.php', 'threatstop.conf.php');
    map { print "adding $_\n"; $source_to_target_conf->{"$template_dir/$_"} = "$target_dir/$_";    } @installable_files;

    &printMsg('DEBUG', "Scheduled files to install: " . Dumper $source_to_target_conf);
    install_config($source_to_target_conf, $search_and_replace, $chown, $chmod);

    update_policy($OPTS);


    return 0;
}

=head3 uninstall_pfsense

1. restore menu hook file
2. restore filter.inc file
3. remove file flags
4. remove php code from /usr/local/www
5. run php disable and remove code
6. flush and purge ACLs

=cut

sub uninstall_pfsense {
    my $OPTS = shift;

    my $template_dir = "$OPTS->{src}/templates/pfsense";
    my $install_base_dir = '/usr/local/www';
    my $hook; # menu hook file
    if (-f "$install_base_dir/head.inc") {
        $hook = 'head.inc';
    } elsif (-f "$install_base_dir/fbegin.inc") {
        $hook = 'fbegin.inc';
    }
    my $hook_path = "$install_base_dir/$hook";
    my $filter_path = '/etc/inc/filter.inc';
    restore_file($hook_path);
    restore_file($filter_path);

    # some telemetry flag / file flag cleanup
    my @files_to_remove = qw(.last_log_upload_attempt .log_upload_status .config_change .last_policy_update .fallback_threatstop_allow .fallback_threatstop_block .policy_download_status .last_log_timestamp);
    for my $fn (@files_to_remove) {
        if (-f "$OPTS->{src}/$fn") {
            &printMsg('DEBUG', "deleting $fn");
            system("$SUDO rm -f $OPTS->{src}/$fn");
        }
    }
    #gui_cfg mods to change mouse-over popup ACL to show only 100 ips
    my $gui_cfg;
    if (-f "$install_base_dir/guiconfig.inc") {
        $gui_cfg = "$install_base_dir/guiconfig.inc";
        restore_file($gui_cfg);
    }

    # this won't work as it requires authenticated php session
    # system("cd $install_base_dir && $SUDO $OPTS->{'php'} ./tsdisable.php > /dev/null 2>&1");
    # if (-f "$OPTS->{src}/tsremove.php") {
    #     system("cd $install_base_dir && $SUDO $OPTS->{'php'} ./tsremove.php");
    # }

    my @installable_files = ('bootstrap-toggle.min.css', 'bootstrap-toggle.min.js', 'ts_logo_bw.png', 'tslogupload.php', 'tsgetblockip.php', 'tsenable.php', 'ts_state.php', 'threatstop.conf.php', 'tssetup.php', 'tsgetblockip.php', 'tsremove.php');
    for my $tsfile (@{installable_files}) {

        if (-f "$install_base_dir/$tsfile") {
            &printMsg('DEBUG', "Found $tsfile file installed, uninstalling..");
            system("$SUDO rm -f $install_base_dir/$tsfile");
        }
    }



    # flush ThreatSTOP tables
    &printMsg('DEBUG', "flushing ThreatSTOP block tables");
    if (system("$SUDO $OPTS->{pfctl} -t ThreatSTOP_block -T flush") == 0 ) {
        &printMsg('OK', "flushed");
    } else {
        &printMsg('FAIL', "flushed");
    }
    &printMsg('DEBUG', "flushing ThreatSTOP allow tables");
    if (system("$SUDO $OPTS->{pfctl} -t ThreatSTOP_allow -T flush") == 0 ) {
        &printMsg('OK', "flushed");
    } else {
        &printMsg('FAIL', "flushed");
    }

    #remove table files
    my $blockdir = '/var/db/aliastables';
    my @policies = ('ThreatSTOP_allow.txt', 'ThreatSTOP_block.txt');
    for my $policy (@policies) {
        if (-f "$blockdir/$policy") {
            if (system("$SUDO rm -f $blockdir/$policy") == 0 ) {
                &printMsg('OK', "deleting $blockdir/$policy");
            } else {
                &printMsg('FAIL', "deleting $blockdir/$policy");
            }
        }
    }


}

1;
