########################################################################
#
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# United States Patent No. 62539504 (Pending),
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#
########################################################################

# This is a Utility library for all things Telemetry
#
# TO USE :
#
# my $t1 = ThreatStop::MultiDevice::Telemetry->new(
#             'tdid_asdf1234',
#             '01',
#             'pan',
#          );
# $t1->genTelemetry ($ttype, $tkey, $tvalue, $severity, $errorref);
# $t1->genTelemetry (...);
# $t1->genTelemetry (...);
# $t1->genTelemetry (...);
#
# $t1->finish(); # <= mandatory for proper cleanup
#
package ThreatStop::Telemetry;
use strict;
use warnings;
use 5.010;    #enable perl 5.10 features.
use Exporter;
use File::Basename;
# use DateTime;
use Data::Dumper;
# use ThreatStop::MultiDevice::Admin;
use Devel::GlobalDestruction;
use FileHandle;                 # to remedy FH has no close method
FileHandle::autoflush STDOUT 1; # to remedy FH has no close method
use ThreatStop;
use ThreatStop::STunnel;
# use Readonly;

# Standard Module defines
our $VERSION   = 0.1;
our @ISA       = qw(Exporter);
our @EXPORT    = qw(genTelemetry $SEVERITIES $TTYPES);
our @EXPORT_OK = qw();

# T_ stands for Telemetry
our $TSDIR = '/opt/threatstop';
our $CFG = {};
our $T_basedir = "$TSDIR/telemetry";
our $T_temp    = "$T_basedir/temp";
our $T_pending = "$T_basedir/pending";
our $T_archive = "$T_basedir/archive";
our $T_suffix  = 'tlm';

# severities
our $SEVERITIES = {
    EMERGENCY => 'emergency',
    ALERT     => 'alert',
    CRITICAL  => 'critical',
    ERROR     => 'error',
    WARNING   => 'warning',
    NOTICE    => 'notice',
    INFO      => 'info',
    DEBUG     => 'debug',
};

# telemetry types
our $TTYPES = {
    PERSISTENT => 'p',
    TRANSIENT => 't',
};

our $FH;
# constructor
#
# a telemetry file is held open for this session
#
# returns :
#   reference to this object
#   upon success, $self->{ERROR} will be undef
#   upon failure, $self->{ERROR} will be set with a reason
#
# args :
#   device ID (TDID)
#   Telemetry format version
#   device type (NOT the particular device name)
#
# Instance variables :
# ERRORS :
#   if setting up the telemetry events file fails,
#   $self->{ERROR} gets set with the reason...see getError()
# FNAME, FHANDLE :
#   the file name & handle to the temporary store in which
#   telemetry events will be written for this instance
#
sub new {

    my $class = shift;
    my $tdid  = shift;
    my $tver  = shift;
    my $dtype = shift;
    my $self = {};
    check_or_create_infrastructure();
    $self->{TDID}    = $tdid;
    $self->{TVER}    = $tver;
    $self->{DTYPE}   = $dtype;
    $self->{ERROR}   = undef;
    $self->{FNAME}   = undef;
    $self->{FHANDLE} = undef;
    $self->{FINISHED}= 0;
    $self->{RECORDS} = []; # holds all, for checksum generation



    my $err = '';
    unless (keys %$CFG) {
        &printMsg('DEBUG', "loading config data");
        &loadConfFileData($ThreatStop::CONFIGFILE, $CFG, \$err);
        # print "config dump: " . Dumper $CFG;
    }

    # my $now = DateTime->now->hires_epoch();
    my $now = time;
    # print "localtime is $now";
    $self->{TIMESTAMP} = $now; # for the file, and every record within
    $self->{FNAME} = "$tdid-$now-$tver.$T_suffix";

    if ( open(FH, ">$T_temp/$self->{FNAME}") ) {

        # FH->autoflush;
        $self->{FHANDLE} = *FH;
        # $self->{FHANDLE} = $FH;
        # $self->{FHANDLE} = *FH;
    } else {

        $self->{ERROR} = $!;
    }
    bless $self, $class;
    return $self;
}

sub check_or_create_infrastructure {
    &printMsg('DEBUG', "creating telemetry infrastructure");
    for my $dir ($T_pending, $T_archive, $T_temp) {
        &printMsg('DEBUG', "Checking on $dir");
        get_or_create_dir($dir);
    }
}

# after instantiation, the caller should check this
# returns
#    undef for "good to go" (ie, no errors), or
#    something that indicates the error
sub getError {

    my $self = shift;
    return $self->{ERROR};
}

# _genZone()
#
# args
# - device id
# - portal telemetry version
#
# returns zone in which to update
# - string : TDID.vVER.telemetry.threatstop.local
sub _genZone {

    my $tdid = shift;
    my $tver = shift;
    my $suffix = 'telemetry.threatstop.local';

    return "$tdid.v$tver.$suffix";
}

# given the passed parameters, this generates a formatted
# telemetry element and writes it to this instances file
# handle
#
# args
# - telemetry type
# - telemetry name
# - telemetry value
# - severity
# - reference to important output that caller may want (like errors)
#
# returns
# 0 for failure
#   - invalid arguments
#   - undef file handle
#   - error from printing to file handle
# 1 for success
sub genTelemetry {

    my $self     = shift;
    my $ttype    = shift;
    my $name     = shift;
    my $value    = shift;
    my $severity = shift;
    my $errorref   = shift;
    if ( ! defined $TTYPES->{$ttype} ) {

        $$errorref .= "Unknown telemetry type : $ttype\n";
        return 0;
    }
    if ( ! defined $SEVERITIES->{$severity} ) {

        $$errorref .= "Unknown severity : $severity\n";
        return 0;
    }

    my $t =
        "v$self->{TVER}." .
        "${name}." .
        "$TTYPES->{$ttype}." .
        "$SEVERITIES->{$severity}." .
        $self->{TIMESTAMP} .
        qq|=$value| ;

    if ( defined $self->{FHANDLE} ) {

        print { $self->{FHANDLE} } "$t\n";
        push @{$self->{RECORDS}}, $t;
    } else {

        $$errorref .= "Undefined telemetry file handle\n";
        return 0;
    }

    return 1;
}

sub _genChecksum {

    my $self = shift;
    my $all = '';
    foreach my $record ( sort @{$self->{RECORDS}} ) {

        $all .= $record;
    }

    return generate_sha1_checksum($CFG, $all);
}

# finish()
# This should be called after all telemetry is loaded
# and the instance is considered complete
#
# add on the meta-records :
# 1. device type
# 2. checksum <= must be last
#
# we close the file handle for this instance and
# move the file to the pending area, ready for
# processing
sub finish {

    my $self = shift;
    my $errors;

    # print 'self: ' . Dumper $self;

    $self->genTelemetry('PERSISTENT', 'device_type', $self->{DTYPE}, 'INFO', \$errors);
    $self->genTelemetry('TRANSIENT', 'checksum', $self->_genChecksum(), 'INFO', \$errors);
    &printMsg('DEBUG', 'Telemetry Generated: ' . Dumper $self->{RECORDS} );


    $self->{FHANDLE}->close() if $self->{FHANDLE};

    # some clients may be running as root, so we must leave the file
    # with threatstop ownership; other processing relies upon it
    my $TSUSER = `getent passwd | awk -F: '\$3 == 1000 { print \$1 }'|tr -d '\n'`;
    if ( $> == 0 ) {

        my ($login,$pass,$uid,$gid) = getpwnam($TSUSER);
        if ( defined $uid and defined $gid ) {

            if ( ! chown($uid, $gid, "$T_temp/$self->{FNAME}") ) {

                $errors .= "unable to chown $T_temp/$self->{FNAME} to $TSUSER";
            }
        } else {

            $errors .= "$TSUSER not in passwd file";
        }
    }

    rename "$T_temp/$self->{FNAME}", "$T_pending/$self->{FNAME}";

    $self->{FINISHED} = 1;
}

# destructor
#
# catch someone forgetting to call finish()
sub DESTROY {

    my $self = shift;

    return if ${^GLOBAL_PHASE} and ${^GLOBAL_PHASE} eq 'DESTRUCT'; # see http://perldoc.perl.org/perlobj.html
    return if in_global_destruction; # see https://metacpan.org/pod/Devel::GlobalDestruction

    $self->finish() if ! $self->{FINISHED};
}

## S T A T I C ---------------
## methods below here are STATIC, ie, NOT called with instance reference

# sendUpdate()
# given the files passed, send the representive content
# via "nsupdate" to the proper telemetry zone
# upon success, the files will be moved to the archive
# directory
#
# note : a single retry will be made upon a detected failure
# of nsupdate
#
# args
# - tdid
# - telemetry version
# - reference to array of file names from the pending directory
# - reference map of tdid to server & port & tsig
# - reference to string in which errors may be place
#
# returns the error string, loaded with something
# helpful, if there were problems
sub sendUpdate {

    my $tdid = shift;
    my $tver = shift;
    my $filesref = shift;
    my $tmap = shift;
    my $errorref = shift;
    my $retried = 0;

    if (stunnel_required($CFG) == 1) {
        $$errorref = 'Failed building Stunnel';
        return;
    }

    AGAIN:
    $$errorref = '';
    my $NSUPDATE_LOG = '/opt/threatstop/telemetry/nsupdate.out';

    # note, the "-v" flag causes TCP to be used
    if ( ! open(NSUPDATE, "|nsupdate -v > $NSUPDATE_LOG 2>&1") ) {

        $$errorref = "Failure to access update utility : $!";
        stunnel_completed($CFG);
        return;
    }

    my $zone = _genZone($tdid, $tver);
    print NSUPDATE "server $tmap->{$tdid}->{ip} $tmap->{$tdid}->{port}\n";
    print NSUPDATE "zone $zone\n";
    &printMsg('DEBUG', "server $tmap->{$tdid}->{ip} $tmap->{$tdid}->{port}");
    &printMsg('DEBUG', "zone $zone");

    foreach my $file ( sort @{$filesref} ) {

        if ( ! open(TFILE, "<$T_pending/$file") ) {

            $$errorref = "Failure : $!";
            stunnel_completed($CFG);
            return;
        }

        my $record;
        while (<TFILE>) {

            chomp;
            print NSUPDATE qq|update add $zone 4242 TXT "$_"\n|;
            &printMsg('DEBUG', "update add $zone 4242 TXT $_");
        }

        close TFILE;
    }
    print NSUPDATE "key $tdid $tmap->{$tdid}->{key}\n";
    print NSUPDATE "send\n";

    &printMsg('DEBUG', "key $tdid $tmap->{$tdid}->{key}");
    &printMsg('DEBUG', "send\n");

    close NSUPDATE;

    if ( $? != 0 ) {

        # any error will leave all files from this attempt in pending
        $$errorref = "Status updates to ThreatSTOP failed";
        chomp ($$errorref .= `cat $NSUPDATE_LOG`);

        &printMsg('ERROR', "$$errorref");

        if ( ! $retried ) {

            sleep 15;
            $retried = 1;
            goto AGAIN;
        }
    } else {

        # no errors detected; archive current file set
        foreach my $file ( @{$filesref} ) {
            &printMsg('DEBUG', "moving $T_pending/$file => $T_archive/$file");
            rename "$T_pending/$file", "$T_archive/$file";
        }
    }
    stunnel_completed($CFG);
}


# sendPending()
# push any pending telemetry to the corresponding
# telemetry zones
#
# Walks the pending directory, processing files that
# look like TDID-timestamp-ver.tlm
#
# args
# - errors ref
#
# It will work on a single device ID at a time, using
#
# returns errors in the passed output reference
sub sendPending {

    my $errorref     = shift; # place errors here

    my $dh;
    &printMsg('DEBUG', "looking in $T_pending for files to send");
    if ( ! opendir($dh, $T_pending) ) {

        $$errorref = "can't read $T_pending: $!";
        return;
    }

    my @pending = readdir($dh);
    closedir $dh;

    # load a map of TDID => { ip => x.x.x.x, port => yyyy, key => tsig }
    # ...will be needed for the actual nsupdate
    my $tdid2server = {};
    my $err = '';
    my $config = $CFG;
    # print "CONFIG dump: " . Dumper $config;


    $tdid2server->{$config->{device_id}}->{ip} = $config->{telemetry_server_ip};
    $tdid2server->{$config->{device_id}}->{port} = $config->{telemetry_server_port};
    $tdid2server->{$config->{device_id}}->{key} = $config->{auto_key};

    my $tdids = {}; # see note below
    my $tdid;
    my $timestamp;
    my $tver;
    foreach my $file ( @pending ) {

        # the files we care about...anthing else
        # (like ., .., cruft), we ignore
        next if $file =~ m/^\./;
        &printMsg('INFO', "checking $file");
        if ( $file !~ /^(tdid_[a-z0-9]{8})-(\d+)-(\d+)\.tlm$/ ) {

            next;
        }
        $tdid = $1;
        $timestamp = $2;
        $tver = $3;

        # re $tdids
        # build up a nice structure from which we can work on all
        # files from the same TDID, remembering that we need the
        # telemetry version in order to target the right telemetry zone
        # $tdids => { tver1 => [ file1, file2, ... ], tver2 => [ file1, file2, ... ], }

        push @{$tdids->{$tdid}->{$tver}}, $file;
    }

    # for each telemetry version in each device ID,
    # push the records on up, and archive what was
    # successfully pushed
    foreach $tdid ( sort keys %{$tdids} ) {

        # it's possible that a device has been deleted, and yet there are still
        # telemetry files pending to be pushed; we have to skip these, as we have
        # no device data for the TDID
        next if ! defined $tdid2server->{$tdid};

        foreach $tver ( sort keys %{$tdids->{$tdid}} ) {

            # (GS-1592) must not overwhelm nsupdate (exceeding DNS package size limit)
            # so build array ref of a limited amount of files, and issue multiple calls
            # to sendUpdate()
            my @allFiles = sort @{$tdids->{$tdid}->{$tver}};
            my @slice;
            my $maxFiles = 50; # a large telemetry file is ~600 bytes...limit max update to ~30kb
            my $lastActualIx = $#allFiles;
            my $lastCandidateIx;
            my $nextIx = 0;
            while ( $nextIx <= $lastActualIx ) {

                $lastCandidateIx = ( $nextIx + $maxFiles - 1  < $lastActualIx ) ? $nextIx + $maxFiles - 1 : $lastActualIx ;
                @slice = @allFiles[ $nextIx .. $lastCandidateIx ];
                sendUpdate($tdid, $tver, \@slice, $tdid2server, $errorref);
                last if $lastCandidateIx >= $lastActualIx;
                $nextIx = ($nextIx + $maxFiles < $lastActualIx) ? $nextIx + $maxFiles : $lastActualIx ;
            }
        }
    }
}

# a simple routine that removes anything in the passed
# directory older than the number passed seconds
#
# args
# - error string reference
#
# returns
# - error string reference loaded with why
#   caller can check if error was modified
sub purgeTelemetry {

    my $pdir = shift;
    my $psecs = shift;
    my $errorref = shift;
    my $dh;
    if ( ! opendir($dh, $pdir) ) {

        $$errorref = "can't read $pdir: $!";
        return;
    }

    my @tfiles = readdir($dh);
    closedir $dh;

    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
        $atime,$mtime,$ctime,$blksize,$blocks);
    # my $now = DateTime->now->hires_epoch();
    my $now = time;
    foreach my $file ( @tfiles ) {

        # the files we care about...anthing else
        # (like ., .., cruft), we ignore
        if ( $file !~ /^(tdid_[a-z0-9]{8})-(\d+)-(\d+)\.tlm$/ ) {

            next;
        }

        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
         $atime,$mtime,$ctime,$blksize,$blocks) = stat("$pdir/$file");

        if ( defined $mtime && $now - $mtime > $psecs ) {

            if ( ! unlink("$pdir/$file") ) {

                $$errorref .= "Fail : $!\n";
            }
        }
    }
}

1;    #End of module
